from iqmotion.clients.client_with_entries import ClientWithEntries
from iqmotion.custom_errors import ClientError

from iqmotion.tests.helpers import make_fake_message

import pytest
import os

from unittest.mock import patch, Mock, MagicMock, call, PropertyMock


class TestClientWithEntries:
    _DICTIONARY_MSG = (["dictionary_client_entry", 1, [1, 3, 99],], 99)
    _PROCESS_MSG = (["process_client_entry", 2, [98], 2], 98)

    @pytest.fixture(params=["client_test"])
    def client(self, request):
        client_file_name = request.param
        test_dir = os.path.dirname(__file__)

        with patch("os.path.dirname") as mock_class:
            mock_class.return_value = test_dir
            client = ClientWithEntries(client_file_name)
            yield client

    def test_client_jsons(self):
        dir_path = os.path.join(os.path.dirname(__file__), "../clients/client_files/")

        for file in os.listdir(dir_path):
            if file.endswith(".json"):
                client_name = file.split(".")[0]
                ClientWithEntries(client_name)

    def test_not_supported_client_entry(self):
        client_file_name = "fail_client_test"
        test_dir = os.path.dirname(__file__)

        with patch("os.path.dirname") as mock_class:
            mock_class.return_value = test_dir

            with pytest.raises(ClientError) as err:
                ClientWithEntries(client_file_name)

        err_str = err.value.message
        assert (
            "CLIENT ERROR: ClientWithEntries does not support this payload type\n"
            == err_str
        )

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG
            # _PROCESS_MSG
        ],
    )
    def test_read_message(self, client, test_input, expected):
        client_entry_name = test_input[0]
        type_idn = test_input[1]
        payload = test_input[2]
        msg = make_fake_message(payload, type_idn)

        client.read_message(msg)

        assert client.get_reply(client_entry_name) == expected

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG
            # _PROCESS_MSG
        ],
    )
    def test_is_fresh(self, client, test_input, expected):
        client_entry_name = test_input[0]
        type_idn = test_input[1]
        payload = test_input[2]
        msg = make_fake_message(payload, type_idn)

        assert client.is_fresh(client_entry_name) == False

        client.read_message(msg)

        assert client.is_fresh(client_entry_name) == True

    def test_default_module_idn(self, client):
        assert client.module_idn == 0

    def test_module_idn(self):
        client_file_name = "client_test"
        test_dir = os.path.dirname(__file__)

        module_idn = 5

        with patch("os.path.dirname") as mock_class:
            mock_class.return_value = test_dir
            client = ClientWithEntries(client_file_name, module_idn)

        assert client.module_idn == module_idn

    @pytest.mark.parametrize(
        "test_input",
        [
            _DICTIONARY_MSG,
            # _PROCESS_MSG
        ],
    )
    def test_client_entries(self, client, test_input):
        client_entry_name = test_input[0][0]
        client_entries = client.client_entries

        assert client_entry_name in client_entries.keys()



import pytest
from iqmotion.communication.circular_queue import CircularQueue


class TestCircularQueue:
    def test_init(self):
        desired_queue_size = 5
        queue = CircularQueue(desired_queue_size)

        assert queue.maxlen == desired_queue_size

    def test_init_from_iterable(self):
        iterable = [1, 2, 3, 4, 5]
        queue = CircularQueue.from_iterable(iterable)

        assert len(queue) == len(iterable)
        assert queue[::] == iterable
        assert type(queue) == CircularQueue

    def test__iter__(self):
        data = [1, 2, 3, 4, 5]
        queue = CircularQueue(5)
        for x in queue:
            # nothing is in queue, should not be reached
            assert 0

        queue.extend(data)
        for ind, x in enumerate(queue):
            assert x == data[ind]

        # test for circularity
        poped_item = queue.popleft()
        queue.append(poped_item)
        poped_item = queue.popleft()
        queue.append(poped_item)
        new_data = [3, 4, 5, 1, 2]

        for ind, x in enumerate(queue):
            assert x == new_data[ind]

    def test__len__(self):
        desired_queue_size = 3
        queue = CircularQueue(desired_queue_size)
        assert len(queue) == 0

        for i in range(3):
            queue.append(i)
            assert len(queue) == i + 1

        queue.append(4)
        assert len(queue) == desired_queue_size

        for i in range(3, 0, -1):
            queue.pop()
            assert len(queue) == i - 1

    def test__str__(self):
        data = [1, 2, 3]
        queue = CircularQueue.from_iterable(data)

        assert data.__str__() == queue.__str__()

    def test__getitem__list(self):
        data = [0, 1, 2, 3]
        queue = CircularQueue.from_iterable(data)
        with pytest.raises(TypeError):
            queue[1.1::]
        with pytest.raises(TypeError):
            queue[:1.1:]
        with pytest.raises(TypeError):
            queue[::1.1]

        for i in range(4):
            assert queue[:i] == data[:i]

        assert queue[2:-1] == data[2:-1]
        assert queue[2:] == data[2:]
        assert queue[::-1] == data[::-1]

    def test__getitem__single(self):
        data = [0, 1, 2, 3]
        queue = CircularQueue.from_iterable(data)

        with pytest.raises(IndexError):
            queue[4]
        with pytest.raises(IndexError):
            queue[-5]

        for i in range(4):
            assert queue[i] == data[i]

        # test for circularity
        queue.popleft()
        queue.append(11)
        queue.popleft()
        queue.append(12)
        new_data = [2, 3, 11, 12]

        for i in range(4):
            assert queue[i] == new_data[i]

    def test_append(self):
        desired_queue_size = 3
        queue = CircularQueue(desired_queue_size)

        for i in range(desired_queue_size):
            assert queue.append(i) == True
            assert queue[-1] == i

        assert queue.append(4) == False

    def test_extend(self):
        desired_queue_size = 3
        data = [0, 1, 2]
        queue = CircularQueue(desired_queue_size)

        queue.extend(data)

        assert queue.extend([1, 2]) == 0

        assert len(queue) == len(data)
        assert queue[::] == data

    def test_pop(self):
        desired_queue_size = 3
        queue = CircularQueue(desired_queue_size)

        assert queue.pop() == None

        for i in range(desired_queue_size):
            queue.append(i)

        for i in range(desired_queue_size - 1, -1, -1):
            assert queue.pop() == i

        assert queue.pop() == None

    def test_popleft(self):
        desired_queue_size = 3
        queue = CircularQueue(desired_queue_size)

        assert queue.popleft() == None

        for i in range(desired_queue_size):
            queue.append(i)

        for i in range(desired_queue_size):
            assert queue.popleft() == i

        assert queue.popleft() == None

    def test_clear(self):
        data = [0, 1, 2, 3]
        queue = CircularQueue.from_iterable(data)

        queue.clear()
        assert len(queue) == 0

    def test_copy(self):
        data = [0, 1, 2, 3]
        queue = CircularQueue.from_iterable(data)

        queue_copy = queue.copy()

        assert queue[::] == queue_copy[::]

    def test_maxlen(self):
        desired_queue_size = 3
        data = [0, 1, 2, 3]
        queue1 = CircularQueue(desired_queue_size)
        queue2 = CircularQueue.from_iterable(data)

        assert queue1.maxlen == desired_queue_size
        assert queue2.maxlen == len(data)

    def test_is_full(self):
        desired_queue_size = 3
        queue = CircularQueue(desired_queue_size)

        for i in range(desired_queue_size):
            assert queue.is_full == False
            queue.append(i)

        assert queue.is_full == True

        for i in range(desired_queue_size):
            queue.pop()
            assert queue.is_full == False

    def test_is_empty(self):
        desired_queue_size = 3
        queue = CircularQueue(desired_queue_size)

        assert queue.is_empty == True

        for i in range(desired_queue_size):
            queue.append(i)
            assert queue.is_empty == False

        for i in range(desired_queue_size):
            queue.pop()

        assert queue.is_empty == True



from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.iq_devices.common_commands.ramper import Ramper


class StepDirModule(IqModule):
    """ Creates StepDirModule object
    
    Arguments:
        com {Communicator} -- The communicator object to interface with the IqModule
    
    Keyword Arguments:
        module_idn {int} -- The idn of the module (default: {0})
    """

    _DEFAULT_CONTROL_CLIENT = "multi_turn_angle_control"
    _MODULE_FILE_NAME = "step_direction.json"

    def ramp_velocity(self, final_velocity: float, total_time: float, time_steps=20):
        """ Ramps velocity of the module up to a target in set amount of seconds

        Arguments:
            final_velocity {float} -- final velocity goal
            total_time {float} -- time to reach velocity goal (s)

        Keyword Arguments:
            time_steps {int} -- num of velocity increments (default: {20})

        Returns:
            bool -- True if the ramp was successful
        """
        velocity_client = self._DEFAULT_CONTROL_CLIENT
        velocity_client_entry = "ctrl_velocity"
        success = Ramper.ramp_velocity(
            self,
            velocity_client,
            velocity_client_entry,
            final_velocity,
            total_time,
            time_steps,
        )

        return success

    def ramp_volts(self, final_volts: float, total_time: float, time_steps=20):
        """ Ramps the volts of the module up to a target in set amount of seconds

        Arguments:
            final_volts {float} -- final volts goal
            total_time {float} -- time to reach volts goal (s)

        Keyword Arguments:
            time_steps {int} -- num of volts increment (default: {20})

        Returns:
            bool -- True if the ramp was successful
        """
        volts_client = self._DEFAULT_CONTROL_CLIENT
        volts_client_entry = "ctrl_volts"
        success = Ramper.ramp_volts(
            self, volts_client, volts_client_entry, final_volts, total_time, time_steps,
        )

        return success

    def ramp_volts_slew(self, final_volts: float, slew_rate: float):
        """ Ramps the volts of the module up to a target following a slew rate

        Arguments:
            final_volts {float} -- final volts goal
            slew_rate {float} -- wanted slew rate

        Returns:
            bool -- True if the ramp was successful
        """
        success = Ramper.ramp_volts_slew(self, final_volts, slew_rate)

        return success



import iqmotion as iq

import pytest


class TestSpeed:
    @pytest.fixture
    def iq_module(self, hardware_type, usb_handle):
        hardware_types = {
            "speed": iq.SpeedModule,
        }

        if hardware_type not in hardware_types.keys():
            print("THIS TYPE IS NOT HANDLED IN TEST SPEED")
            exit()

        com = iq.SerialCommunicator(usb_handle)
        iq_module = hardware_types[hardware_type](com, 0)

        # timeout for speed modules
        if hardware_type == "speed":
            iq_module.set("propeller_motor_control", "timeout", 10)

        return iq_module

    def test_ramp_velocity(self, iq_module):
        assert iq_module.ramp_velocity(50, 1)
        assert iq_module.ramp_velocity(-50, 2)
        assert iq_module.ramp_velocity(0, 1)

    def test_ramp_volts(self, iq_module):
        assert iq_module.ramp_volts(0.5, 1)
        assert iq_module.ramp_volts(-0.5, 2)
        assert iq_module.ramp_volts(0, 1)

    def test_ramp_volts_slew(self, iq_module):
        assert iq_module.ramp_volts_slew(0.5, 0.5)
        assert iq_module.ramp_volts_slew(-0.5, 0.5)
        assert iq_module.ramp_volts_slew(0, 0.5)



from iqmotion.message_making.dictionary_message_maker import DictionaryMessageMaker
from iqmotion.message_making.dictionary_message_maker import DictionaryMessageMakerData
from iqmotion.client_entries.dictionary_client_entry import DictionaryClientEntry
from iqmotion.client_entries.dictionary_client_entry import AccessType
from iqmotion.custom_errors import MessageMakerError


import pytest


class TestDictionaryMessageMaker:
    _FAKE_DICTIONARY_CLIENT_ENTRY_DATA_DICT = {
        "type_idn": 1,
        "param": "dictionary_client_entry",
        "param_idn": 1,
        "format": "B",
        "unit": "rad",
    }

    _FAKE_DICTIONARY_CLIENT_ENTRY_DATA_DICT_LIST = {
        "type_idn": 1,
        "param": "dictionary_client_entry",
        "param_idn": 1,
        "format": "BB*",
        "unit": "rad",
    }

    _TEST_VALUES_ACCESSTYPES = [
        ([AccessType.GET, 0], [1, 1, 0, 0]),
        ([AccessType.SET, 1], [1, 1, 1, 1]),
        ([AccessType.SAVE, 2], [1, 1, 2, 2]),
        ([AccessType.REPLY, 3], [1, 1, 3, 3]),
    ]

    _TEST_VALUES_ACCESSTYPES_LIST = [
        ([AccessType.GET, [0, 0]], [1, 1, 0, 0, 0]),
        ([AccessType.SET, [1, 1]], [1, 1, 1, 1, 1]),
        ([AccessType.SAVE, [2, 2, 2]], [1, 1, 2, 2, 2, 2]),
        ([AccessType.REPLY, [3, 3, 3, 3]], [1, 1, 3, 3, 3, 3, 3]),
    ]

    @pytest.fixture
    def test_input(self):
        """dummy fixture that changes with the pytest.mark.parametrize
        in the actual tests
        """
        return ([AccessType.GET, 0], 0)

    @pytest.fixture()
    def message_maker(self, test_input):
        client_entry = DictionaryClientEntry(
            self._FAKE_DICTIONARY_CLIENT_ENTRY_DATA_DICT
        )

        acces_type = test_input[0]
        value = test_input[1]

        message_maker = DictionaryMessageMaker(client_entry, 0, acces_type, value)

        return message_maker

    @pytest.fixture()
    def message_maker_with_list_client_entry(self, test_input):
        client_entry = DictionaryClientEntry(
            self._FAKE_DICTIONARY_CLIENT_ENTRY_DATA_DICT_LIST
        )

        acces_type = test_input[0]
        value = test_input[1]

        message_maker = DictionaryMessageMaker(client_entry, 0, acces_type, value)

        return message_maker

    @pytest.mark.parametrize("test_input, expected", _TEST_VALUES_ACCESSTYPES)
    def test_make(self, message_maker, test_input, expected):
        message = message_maker.make()

        assert message == bytearray(expected)

    @pytest.mark.parametrize("test_input, expected", _TEST_VALUES_ACCESSTYPES_LIST)
    def test_make_with_list_values(
        self, message_maker_with_list_client_entry, test_input, expected
    ):
        message = message_maker_with_list_client_entry.make()

        assert message == bytearray(expected)

    @pytest.mark.parametrize("test_input, expected", _TEST_VALUES_ACCESSTYPES_LIST)
    def test_make_values_too_long(self, message_maker, test_input, expected):

        with pytest.raises(MessageMakerError) as err:
            message_maker.make()

        err_str = err.value.message
        assert "MESSAGE MAKER ERROR: values too long for client entry\n" == err_str



from iqmotion.clients.client import Client
from iqmotion.custom_errors import ClientError
from iqmotion.client_entries.dictionary_client_entry import DictionaryClientEntry

# from iqmotion.communication.process_client_entry import ProcessClientEntry

import os
import json


class ClientWithEntries(Client):
    """ ClientWithEntries is an implementation of Client
        A ClientWithEntries object is able to read a message to store its content if needed, check if it's fresh and retrieve it.

        A ClientWithEntries is defined by a .json file where all its entries data is located
    """

    def __init__(self, client_file_name: str, module_idn=0):
        self._client_entry_dict = {}
        self._module_idn = module_idn

        client_file = self._parse_client_json(client_file_name)

        self._populate_client_entries(client_file)

    def _parse_client_json(self, client_file_name: str):
        client_json = client_file_name + ".json"

        file_path = os.path.join(
            os.path.dirname(__file__), ("client_files/" + client_json)
        )

        with open(file_path) as json_file:
            client_file = json.load(json_file)

        return client_file

    def _populate_client_entries(self, client_file: dict):
        for client_entry_data_dict in client_file:
            client_entry_name = client_entry_data_dict["param"]
            client_entry = self._create_client_entry(client_entry_data_dict)

            self._client_entry_dict[client_entry_name] = client_entry

        return

    def _create_client_entry(self, client_entry_data_dict: dict):
        # special case where no "payload_type" field exists, legacy compatibility
        if "payload_type" not in client_entry_data_dict.keys():
            client_entry = DictionaryClientEntry(client_entry_data_dict)
        else:
            raise ClientError("ClientWithEntries does not support this payload type")

        # UNCOMMENT WHEN YOU HANDLE PROCESS CLIENT ENTRY
        # payload_type = client_entry_data_dict["payload_type"]
        # if payload_type == 1:
        #     client_entry = ProcessClientEntry(client_entry_data_dict)
        # else:
        #     raise ClientError(
        #         "ClientWithEntries does not support this payload type")

        return client_entry

    def read_message(self, message: bytearray):
        """ Takes in a message and puts it in the right client entries by matching type_idns
        """
        msg_type_idn = message[0]

        for client_entry in self._client_entry_dict.values():
            type_idn = client_entry.data.type_idn

            if type_idn == msg_type_idn:
                client_entry.read_message(message)

    def is_fresh(self, value_name: str = ""):
        """ Checks if a specific client entry has a fresh value

        Args:
            value_name (String): Client entry name

        Returns:
            True: if value is fresh
            False: if value is not fres
        """
        client_entry = self._client_entry_dict[value_name]
        return client_entry.fresh

    def get_reply(self, value_name: str = ""):
        """ Gets the value from a specific client entry

        Args:
            value_name (String): Client entry name

        Returns:
            (format): value from the client entry with its type define by the format entry
        """
        client_entry = self._client_entry_dict[value_name]
        return client_entry.value

    @property
    def module_idn(self):
        """ Returns the module idn of the client

        Returns:
            int: module idn
        """
        return self._module_idn

    @property
    def client_entries(self):
        """ Returns a dictionary of client entries available in this client

        Returns:
            dict: dictonary of client entries {client_entry_name(String): ClienEntry}
        """
        return self._client_entry_dict



import numpy as np


class Crc:
    """ A static class that creates a (uint16) CRC value from some list of bytes
    """

    @staticmethod
    def make_crc(data):
        """ Make a crc from a list of bytes

        Args:
            data (List): a list of bytes to make the crc from

        Returns:
            crc (np.uint16): a uint16 CRC
        """
        crc = np.uint16(0xFFFF)

        for i in range(len(data)):
            crc = Crc.byte_update_crc(crc, data[i])
        pass

        return crc

    @staticmethod
    def byte_update_crc(crc, data: bytes):
        """ Update a current crc with a new byte 

        Args:
            crc (np.uint16): current crc value
            data (bytes): the byte used to update the crc

        Returns:
            crc (np.uint16): new updated crc
        """
        x = np.uint16((crc >> 8) ^ data)
        x ^= np.uint16(x >> 4)

        crc = np.uint16((crc << 8) ^ (x << 12) ^ (x << 5) ^ x)
        return crc



from iqmotion.communication.serial_packet_states import SerialStartState
from iqmotion.communication.serial_packet_states import SerialLenState
from iqmotion.communication.serial_packet_states import SerialTypeState
from iqmotion.communication.serial_packet_states import SerialPayloadState
from iqmotion.communication.serial_packet_states import SerialCrcState
from iqmotion.communication.crc import Crc
from iqmotion.communication.circular_queue import CircularQueue
from iqmotion.custom_errors import PacketStateError


from iqmotion.tests.helpers import make_fake_packet

import pytest


@pytest.fixture(params=[[1, 2, 3, 4, 5], []])
def circular_queue_with_packet(request):
    packet = make_fake_packet(request.param)
    cq = CircularQueue.from_iterable(packet)

    return [cq, packet]


class TestSerialStartState:
    def test_no_packet(self):
        data = [1, 2, 3, 4, 5]
        cq = CircularQueue.from_iterable(data)

        state = SerialStartState(cq)

        state.parse()

        assert state.find_next_state() == state

    def test_packet(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]

        state = SerialStartState(cq)

        state.parse()

        assert type(state.find_next_state()) == SerialLenState


class TestSerialLenState:
    def test_wrong_parse_index(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]
        len_data = len(circular_queue_with_packet[1])
        parse_index = len_data + 1

        state = SerialLenState(cq, 0, parse_index, 0)
        state.parse()

        assert state.find_next_state() == state

    def test_packet(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]

        state = SerialStartState(cq)
        state.parse()
        state = state.find_next_state()

        state.parse()

        assert type(state.find_next_state()) == SerialTypeState

    def test_overflow_packet(self):
        packet = []
        for i in range(256):
            packet.append(i)

        packet[1] = 256

        cq = CircularQueue.from_iterable(packet)
        state = SerialLenState(cq)

        with pytest.raises(PacketStateError) as err:
            state.parse()

        err_str = err.value.message
        assert (
            "PACKET STATE ERROR: Packet overflow, message is bigger than 256 bytes\n"
            == err_str
        )


class TestSerialTypeState:
    def test_wrong_parse_index(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]
        len_data = len(circular_queue_with_packet[1])
        parse_index = len_data + 1

        state = SerialTypeState(cq, 0, parse_index, 0)
        state.parse()

        assert state.find_next_state() == state

    def test_packet(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]
        payload_len = circular_queue_with_packet[1][1]

        state = SerialStartState(cq)
        state.parse()

        for _ in range(2):
            state = state.find_next_state()
            state.parse()

        new_state_type = type(state.find_next_state())

        if payload_len == 0:
            assert new_state_type == SerialCrcState
        else:
            assert new_state_type == SerialPayloadState


class TestSerialPayloadState:
    def test_wrong_parse_index(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]
        len_data = len(circular_queue_with_packet[1])
        parse_index = len_data + 1

        state = SerialPayloadState(cq, 0, parse_index, 5)
        state.parse()

        assert state.find_next_state() == state

    def test_packet(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]

        state = SerialStartState(cq)
        state.parse()

        for _ in range(3):
            state = state.find_next_state()
            state.parse()

        assert type(state.find_next_state()) == SerialCrcState


class TestSerialCrcState:
    def test_no_crc(self):
        data = [1, 2, 3, 4, 5]
        cq = CircularQueue.from_iterable(data)

        state = SerialCrcState(cq)
        state.parse()

        assert state.is_succesful == False

    def test_crc_state_packet(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]
        message = circular_queue_with_packet[1][2:-2]

        state = SerialStartState(cq)
        state.parse()

        for _ in range(4):
            state = state.find_next_state()
            state.parse()

        assert state.message == message

    def test_wrong_parse_index(self, circular_queue_with_packet):
        cq = circular_queue_with_packet[0]
        len_data = len(circular_queue_with_packet[1])
        parse_index = len_data + 1

        state = SerialCrcState(cq, 0, parse_index, 0)
        state.parse()

        assert state.is_succesful == False



from iqmotion.communication.serial_communicator import SerialCommunicator
from iqmotion.communication.crc import Crc
from iqmotion.custom_errors import CommunicationError

from iqmotion.tests.helpers import DummySerial
from iqmotion.tests.helpers import make_fake_packet

import pytest
import serial

from unittest.mock import patch, Mock, MagicMock, call, PropertyMock


class TestSerialCommunicator:
    @pytest.fixture
    def mock_serial(self, monkeypatch):
        def mock_find_serial_port(*args, **kwargs):
            return ["test_port", "test_port2"]

        monkeypatch.setattr(
            SerialCommunicator, "_find_serial_port", mock_find_serial_port
        )

        with patch("serial.Serial") as mock_class:
            mock_class.return_value = DummySerial("test_port", 115200)
            mock_serial = mock_class.return_value

            yield mock_serial

    def test__init__wrong_port(self, mock_serial):
        with pytest.raises(CommunicationError) as err:
            SerialCommunicator("wrong_port")

        pretty_available_ports_str = '\t"test_port"\n\t"test_port2"\n'
        expected_err = (
            "COMMUNICATION ERROR: Serial port is not available, here is a list of available ports:\n"
            + pretty_available_ports_str
        )

        err_str = err.value.message
        assert expected_err == err_str

    def test__del__(self, mock_serial):
        mock_serial.close = MagicMock()
        ser = SerialCommunicator("test_port")
        del ser

        assert mock_serial.close.called == True

    def test_send_now(self, mock_serial):
        mock_serial.write = MagicMock()
        ser = SerialCommunicator("test_port")

        packet = bytearray([1, 2, 3, 4])
        ser.add_to_out_queue(packet)
        ser.send_now()
        assert mock_serial.write.call_args == call(packet)

    def test_send_now_empty_out_queue(self, mock_serial):
        mock_serial.write = MagicMock()
        ser = SerialCommunicator("test_port")

        ser.send_now()
        assert mock_serial.write.call_args == call(bytearray([]))

    def test_send_message(self, mock_serial):
        mock_serial.write = MagicMock()
        ser = SerialCommunicator("test_port")

        message = bytearray([0, 11, 12, 13, 14])
        ser.send_message(message)

        packet = make_fake_packet(message[1:], 0)

        assert mock_serial.write.call_args == call(packet)

    def test_flush_input_buffer(self, mock_serial):
        # fake_message = bytearray([i for i in range(100)])
        packet = bytes([0, 1, 2, 3, 4])
        mock_serial.reset_input_buffer = MagicMock()
        mock_serial.read = MagicMock(return_value=packet)
        mock_serial._in_waiting = len(packet)

        ser = SerialCommunicator("test_port")
        ser.read_bytes()
        assert ser.bytes_left_in_queue == True

        ser.flush_input_buffer()

        assert mock_serial.reset_input_buffer.called == True
        assert ser.bytes_left_in_queue == False

    @pytest.mark.parametrize(
        "test_input", [(bytes([0, 1, 2, 3, 4])), (bytearray([i for i in range(255)]))],
    )
    def test_read_bytes(self, mock_serial, test_input):
        packet = test_input
        mock_serial.read = MagicMock(return_value=packet)
        mock_serial._in_waiting = len(packet)
        ser = SerialCommunicator("test_port")
        read_all_available = ser.read_bytes()

        assert mock_serial.read.call_args == call(len(packet))
        assert read_all_available == True

    def test_read_bytes_overflow(self, mock_serial):
        packet = bytes([10 for i in range(600)])
        mock_serial.read = MagicMock(return_value=packet[: (255 + 5) * 2])
        mock_serial._in_waiting = len(packet)

        ser = SerialCommunicator("test_port")
        read_all_available = ser.read_bytes()

        assert mock_serial.read.call_args == call((255 + 5) * 2)
        assert read_all_available == False

    @pytest.mark.parametrize(
        "test_input", [(bytes([0, 1, 2, 3, 4])), (bytearray([i for i in range(255)]))],
    )
    def test_byte_left_in_queue(self, mock_serial, test_input):
        packet = test_input
        mock_serial.read = MagicMock(return_value=packet)
        mock_serial._in_waiting = len(packet)
        ser = SerialCommunicator("test_port")
        ser.read_bytes()

        assert ser.bytes_left_in_queue == True

    def test_byte_left_in_queue_empty(self, mock_serial):
        ser = SerialCommunicator("test_port")

        assert ser.bytes_left_in_queue == False

    @pytest.mark.parametrize(
        "test_input", [(bytes([0, 1, 2, 3, 4])), (bytearray([i for i in range(255)]))],
    )
    def test_extract_message(self, mock_serial, test_input):
        ser = SerialCommunicator("test_port")

        message = test_input
        packet = bytes(make_fake_packet(message, 0))

        mock_serial.read = MagicMock(return_value=packet)
        mock_serial._in_waiting = len(packet)
        ser.read_bytes()

        message_back = ser.extract_message()

        assert message_back[1:] == message

    def test_extract_message_from_empty_queue(self, mock_serial):
        ser = SerialCommunicator("test_port")

        message_back = ser.extract_message()

        assert message_back == None




from iqmotion.iq_devices.speed_module import SpeedModule
from iqmotion.iq_devices.servo_module import ServoModule
from iqmotion.iq_devices.step_dir_module import StepDirModule

from iqmotion.communication.serial_communicator import SerialCommunicator



from iqmotion.clients import client_with_entries
from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.iq_devices.iq_module_json_parser import IqModuleJsonParser
from iqmotion.communication.communicator import Communicator
from iqmotion.custom_errors import IqModuleError

from iqmotion.tests.helpers import MockCommunicator

import os
import json
import serial

import pytest
from unittest.mock import patch, Mock, MagicMock, call, PropertyMock


class TestIqModule:
    _CLIENT_NAME = "client_test"

    _MODULE_FILE_NAME = "iq2306_2200kv.json"

    _DICTIONARY_MSG_SET = (["dictionary_client_entry", 99], [1, 1, 1, 99])
    _EMPTY_DICTIONARY_MSG_SET = (
        ["dictionary_client_entry_no_format", []],
        [3, 3, 1],
    )

    _PROCESS_MSG_SET = (["process_client_entry", 98], [2, 2, 1, 98])

    _DICTIONARY_MSG_GET = ("dictionary_client_entry", [1, 1, 0])

    _DICTIONARY_MSG_SAVE = ("dictionary_client_entry", [1, 1, 2])

    _DICTIONARY_MSG_REPLY = (["dictionary_client_entry", [1, 1, 3, 99]], 99)

    @pytest.fixture
    def mock_communicator(self):
        mock_class = MockCommunicator()
        return mock_class

    @pytest.fixture
    def mock_client(self, request):
        client_file_name = self._CLIENT_NAME
        test_dir = os.path.dirname(__file__)

        with patch("os.path.dirname") as mock_class:
            mock_class.return_value = test_dir
            test_client = client_with_entries.ClientWithEntries(client_file_name)

        with patch(
            "iqmotion.clients.client_with_entries.ClientWithEntries"
        ) as mock_class:
            with patch.object(IqModuleJsonParser, "parse") as mock_parser:
                mock_parser.return_value = {"clients": [self._CLIENT_NAME]}
                mock_class.return_value = test_client
                mock_client = mock_class.return_value
                yield mock_client

    def test_wrong_client(self, mock_communicator, mock_client):
        module = IqModule(mock_communicator)
        client_name = "fake_client"
        client_entry_name = "fake_client_entry"

        with pytest.raises(IqModuleError) as err:
            module.set(client_name, client_entry_name, 0)

        err_str = err.value.message
        assert (
            "IQ MODULE ERROR: This module does not support this client: {0}\n\n".format(
                client_name
            )
            == err_str
        )

    def test_wrong_client_entry(self, mock_communicator, mock_client):
        module = IqModule(mock_communicator)
        client_name = self._CLIENT_NAME
        client_entry_name = "fake_client_entry"

        with pytest.raises(IqModuleError) as err:
            module.set(client_name, client_entry_name, 0)

        err_str = err.value.message
        assert (
            "IQ MODULE ERROR: {0} does not support this client entry: {1}\n\n".format(
                client_name, client_entry_name
            )
            == err_str
        )

    def test_coast(self, mock_communicator, mock_client):
        module = IqModule(mock_communicator)
        module.set = MagicMock()
        module.coast()

        assert module.set.called

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_SET,
            _EMPTY_DICTIONARY_MSG_SET,
            # _PROCESS_MSG_SET
        ],
    )
    def test_set(self, mock_communicator, mock_client, test_input, expected):
        mock_communicator.send_message = MagicMock()
        module = IqModule(mock_communicator)

        client_entry_name = test_input[0]
        value = test_input[1]
        module.set(self._CLIENT_NAME, client_entry_name, value)

        assert mock_communicator.send_message.call_args == call(bytearray(expected))

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_SET,
            _EMPTY_DICTIONARY_MSG_SET,
            # _PROCESS_MSG_SET
        ],
    )
    def test_set_retry(self, mock_communicator, mock_client, test_input, expected):
        mock_communicator.send_message = MagicMock()
        module = IqModule(mock_communicator)
        module.get_retry = MagicMock()

        client_entry_name = test_input[0]
        value = test_input[1]
        if not value:
            module.get_retry.return_value = 1
            module.set_verify(self._CLIENT_NAME, client_entry_name)
        else:
            module.get_retry.return_value = value
            module.set_verify(self._CLIENT_NAME, client_entry_name, value)

        assert module.get_retry.call_args == call(
            self._CLIENT_NAME, client_entry_name, 0.1, 5
        )

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_SET,
            _EMPTY_DICTIONARY_MSG_SET,
            # _PROCESS_MSG_SET
        ],
    )
    def test_set_retry_fail(self, mock_communicator, mock_client, test_input, expected):
        mock_communicator.send_message = MagicMock()
        module = IqModule(mock_communicator)
        module.get_retry = MagicMock()

        client_entry_name = test_input[0]
        value = test_input[1]
        if not value:
            module.get_retry.return_value = 0
            success = module.set_verify(self._CLIENT_NAME, client_entry_name)
        else:
            module.get_retry.return_value = None
            success = module.set_verify(self._CLIENT_NAME, client_entry_name, value)
            # two points of failure
            assert success == False

            module.get_retry.return_value = value - 1
            success = module.set_verify(self._CLIENT_NAME, client_entry_name, value)

        assert success == False

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_SET,
            _EMPTY_DICTIONARY_MSG_SET,
            # _PROCESS_MSG_SET
        ],
    )
    def test_set_retry_save(self, mock_communicator, mock_client, test_input, expected):
        mock_communicator.send_message = MagicMock()
        module = IqModule(mock_communicator)
        module.save = MagicMock()
        module.get_retry = MagicMock()

        client_entry_name = test_input[0]
        value = test_input[1]
        if not value:
            module.get_retry.return_value = 1
            module.set_verify(self._CLIENT_NAME, client_entry_name, save=1)
        else:
            module.get_retry.return_value = value
            module.set_verify(self._CLIENT_NAME, client_entry_name, value, save=1)

        assert module.save.call_args == call(self._CLIENT_NAME, client_entry_name)

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_REPLY,
            # _PROCESS_MSG_REPLY
        ],
    )
    def test_get(self, mock_communicator, mock_client, test_input, expected):
        client_entry_name = test_input[0]
        valid_message = test_input[1]

        mock_communicator.read_bytes = MagicMock()
        mock_communicator.extract_message = MagicMock()
        mock_communicator.extract_message.side_effect = [bytearray(valid_message), None]

        module = IqModule(mock_communicator)

        assert module.get("client_test", client_entry_name) == expected

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_REPLY,
            # _PROCESS_MSG_REPLY
        ],
    )
    def test_get_retry(self, mock_communicator, mock_client, test_input, expected):
        client_entry_name = test_input[0]
        valid_message = test_input[1]

        mock_communicator.read_bytes = MagicMock()
        mock_communicator.extract_message = MagicMock()
        mock_communicator.extract_message.side_effect = [bytearray(valid_message), None]

        module = IqModule(mock_communicator)

        retries = 5
        reply = module.get_retry("client_test", client_entry_name, retries=retries)

        assert reply == expected

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_REPLY,
            # _PROCESS_MSG_REPLY
        ],
    )
    def test_get_retry_fail(self, mock_communicator, mock_client, test_input, expected):
        client_entry_name = test_input[0]

        module = IqModule(mock_communicator)

        module.get = MagicMock()
        module.get.return_value = None

        retries = 5
        _ = module.get_retry("client_test", client_entry_name, retries=retries)

        assert module.get.call_count == retries

    @pytest.mark.parametrize(
        "test_input",
        [
            _DICTIONARY_MSG_REPLY,
            # _PROCESS_MSG_GET
        ],
    )
    def test_get_timeout(self, mock_communicator, mock_client, test_input):
        client_entry_name = test_input[0][0]
        module = IqModule(mock_communicator)

        assert module.get("client_test", client_entry_name) == None

    def test_get_all(self, mock_communicator, mock_client):
        module = IqModule(mock_communicator)
        module.get = MagicMock()
        module.get.return_value = "test"
        num_client_entries = len(mock_client.client_entries)

        module.get_all(self._CLIENT_NAME)

        assert module.get.call_count == num_client_entries

    def test_get_all_retry(self, mock_communicator, mock_client):
        module = IqModule(mock_communicator)
        module.get_retry = MagicMock()
        module.get_retry.return_value = "test"
        num_client_entries = len(mock_client.client_entries)

        module.get_all_retry(self._CLIENT_NAME)

        assert module.get_retry.call_count == num_client_entries

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_SAVE,
            # _PROCESS_MSG_SAVE
        ],
    )
    def test_save(self, mock_communicator, mock_client, test_input, expected):
        mock_communicator.send_message = MagicMock()
        module = IqModule(mock_communicator)

        client_entry_name = test_input
        module.save(self._CLIENT_NAME, client_entry_name)

        assert mock_communicator.send_message.call_args == call(bytearray(expected))

    def test_save_all(self, mock_communicator, mock_client):
        module = IqModule(mock_communicator)
        module.save = MagicMock()
        module.save.return_value = "test"
        num_client_entries = len(mock_client.client_entries)

        module.save_all(self._CLIENT_NAME)

        assert module.save.call_count == num_client_entries

    def test_flush_input_buffer(self, mock_communicator, mock_client):
        mock_communicator.flush_input_buffer = MagicMock()
        module = IqModule(mock_communicator)
        module.flush_input_com_buffer()

        assert mock_communicator.flush_input_buffer.called == True

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_GET,
            # _PROCESS_MSG_GET
        ],
    )
    def test_get_async(self, mock_communicator, mock_client, test_input, expected):
        mock_communicator.send_message = MagicMock()
        module = IqModule(mock_communicator)

        client_entry_name = test_input
        module.get_async(self._CLIENT_NAME, client_entry_name)

        assert mock_communicator.send_message.call_args == call(bytearray(expected))

    def test_update_replies(self, mock_communicator, mock_client):
        mock_communicator.read_bytes = MagicMock()
        mock_communicator.extract_message = MagicMock()
        mock_communicator.extract_message.side_effect = [bytearray([1, 2, 3]), None]

        module = IqModule(mock_communicator)
        module.update_replies()

        assert mock_communicator.read_bytes.call_count == 1
        assert mock_communicator.extract_message.called == True

    def test_update_reply(self, mock_communicator, mock_client):
        mock_communicator.read_bytes = MagicMock()
        mock_communicator.extract_message = MagicMock()
        mock_communicator.extract_message.side_effect = [None, bytearray([1, 2, 3])]

        module = IqModule(mock_communicator)
        module.update_reply()

        assert mock_communicator.read_bytes.call_count == 1
        assert mock_communicator.extract_message.call_count == 2

    @pytest.mark.parametrize(
        "test_input",
        [
            _DICTIONARY_MSG_REPLY,
            # _PROCESS_MSG_REPLY
        ],
    )
    def test_is_fresh(self, mock_communicator, mock_client, test_input):
        client_entry_name = test_input[0][0]
        valid_message = test_input[0][1]

        mock_communicator.read_bytes = MagicMock()
        mock_communicator.extract_message = MagicMock()
        mock_communicator.extract_message.side_effect = [bytearray(valid_message), None]

        module = IqModule(mock_communicator)
        assert module.is_fresh(self._CLIENT_NAME, client_entry_name) == False
        module.update_replies()
        assert module.is_fresh(self._CLIENT_NAME, client_entry_name) == True

    @pytest.mark.parametrize(
        "test_input,expected",
        [
            _DICTIONARY_MSG_REPLY,
            # _PROCESS_MSG_REPLY
        ],
    )
    def test_get_reply(self, mock_communicator, mock_client, test_input, expected):
        client_entry_name = test_input[0]
        valid_message = test_input[1]

        mock_communicator.read_bytes = MagicMock()
        mock_communicator.extract_message = MagicMock()
        mock_communicator.extract_message.side_effect = [bytearray(valid_message), None]

        module = IqModule(mock_communicator)
        module.update_replies()

        assert module.get_reply("client_test", client_entry_name) == expected



import pytest

from iqmotion.communication.crc import Crc


class TestCrc:
    def test_make_crc(self):
        expected_crc = [61905, 3708, 44461]
        data = [1, 2, 3]
        count = 3
        crcs = []
        for i in range(count):
            crcs.append(Crc.make_crc(data[: i + 1]))

        for i in range(count):
            assert crcs[i] == expected_crc[i]

    def test_updateCrc(self):
        crc = 0xFFFF
        expected_crc = [61905, 49586, 53651]
        data = [1, 2, 3]
        crcs = []

        for i in range(len(data)):
            crcs.append(Crc.byte_update_crc(crc, data[i]))

        for i in range(len(data)):
            assert crcs[i] == expected_crc[i]



from iqmotion.communication.crc import Crc


class MockCommunicator:
    def send_message(self, message: bytearray):
        return

    def add_to_out_queue(self, bytes: bytearray):
        return

    def send_now(self):
        return

    def read_bytes(self):
        return

    def flush_input_buffer(self):
        return

    @property
    def bytes_left_in_queue(self):
        return 0

    def extract_message(self):
        return


class DummySerial:
    def __init__(self, port, baudrate):
        self._port = port
        self._baudrate = baudrate
        self._is_open = True
        self._in_waiting = 0

    def write(self, input_data):
        return

    def read(self, num_bytes):
        return

    def close(self):
        return

    def reset_input_buffer(self):
        return

    @property
    def in_waiting(self):
        return self._in_waiting

    @property
    def is_open(self):
        return self._is_open


def make_fake_packet(data, type_idn=0):
    data_type = type_idn
    data_len = len(data)

    start = 0x55

    crc_data = [data_len, data_type]
    crc_data.extend(data)
    crc = Crc.make_crc(crc_data)
    crcl = crc & 0xFF
    crch = crc >> 8

    packet = [start, data_len, data_type]
    packet.extend(data)
    packet.append(crcl)
    packet.append(crch)

    return bytearray(packet)


def make_fake_message(data, type_idn=0):
    data_type = type_idn
    message = bytearray([data_type])
    message.extend(data)

    return message



def pytest_addoption(parser):
    parser.addoption("--hardware_type", action="store", default="speed")
    parser.addoption("--usb_handle", action="store", default="/dev/ttyUSB0")


def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".
    option_value = metafunc.config.option.hardware_type
    if "hardware_type" in metafunc.fixturenames and option_value is not None:
        metafunc.parametrize("hardware_type", [option_value])

    option_value = metafunc.config.option.usb_handle
    if "usb_handle" in metafunc.fixturenames and option_value is not None:
        metafunc.parametrize("usb_handle", [option_value])



from iqmotion.client_entries.dictionary_client_entry import DictionaryClientEntry
from iqmotion.client_entries.dictionary_client_entry import DictionaryClientEntryData

import pytest

TEST_CLIENT_ENTRY_DICT = {
    "type_idn": 50,
    "param": "test",
    "param_idn": 0,
    "format": "B",
    "unit": "fake",
}


class TestDictionaryClientEntry:
    def test_read_message(self):
        message = bytearray([50, 0, 3, 99])
        dce = DictionaryClientEntry(TEST_CLIENT_ENTRY_DICT)

        dce.read_message(message)

        assert dce.fresh == True

    def test_read_message_wrong_access_type(self):
        message = bytearray([50, 0, 1, 99])
        dce = DictionaryClientEntry(TEST_CLIENT_ENTRY_DICT)

        dce.read_message(message)

        assert dce.fresh == False

    def test_fresh(self):
        dce = DictionaryClientEntry(TEST_CLIENT_ENTRY_DICT)

        assert dce.fresh == False

        dce.value = bytearray([11])
        assert dce.fresh == True

        _ = dce.value
        assert dce.fresh == False

    def test_value(self):
        dce = DictionaryClientEntry(TEST_CLIENT_ENTRY_DICT)

        assert dce.value == None

        dce.value = bytearray([11])
        assert dce.value == 11

    def test_data(self):
        dce = DictionaryClientEntry(TEST_CLIENT_ENTRY_DICT)

        expected_data = DictionaryClientEntryData(TEST_CLIENT_ENTRY_DICT)

        assert dce.data == expected_data

    def test_data__str__(self):
        dce = DictionaryClientEntry(TEST_CLIENT_ENTRY_DICT)

        expected_str = "{0:10} | {1:2}: {2:32} {3:4} {4}".format(
            TEST_CLIENT_ENTRY_DICT["type_idn"],
            TEST_CLIENT_ENTRY_DICT["param_idn"],
            TEST_CLIENT_ENTRY_DICT["param"],
            TEST_CLIENT_ENTRY_DICT["format"],
            TEST_CLIENT_ENTRY_DICT["unit"],
        )

        assert dce.data.__str__() == expected_str



from iqmotion.iq_devices.iq_module import IqModule

import numpy as np
import time


class Ramper:
    @staticmethod
    def ramp_velocity(
        iq_module: IqModule,
        velocity_client: str,
        velocity_client_entry: str,
        final_velocity: float,
        total_time: float,
        time_steps=20,
    ):
        init_velocity = iq_module.get_retry(
            "brushless_drive", "obs_velocity", retries=5
        )

        if init_velocity == None:
            return 0

        num_steps = round(1000 * total_time / time_steps)  # 20ms time steps by default

        velocity_steps = np.linspace(init_velocity, final_velocity, num_steps)
        time_steps = np.linspace(0, total_time, num_steps)

        start_time = time.perf_counter()
        for step in range(num_steps):
            while time.perf_counter() < start_time + time_steps[step]:
                pass
            new_velocity = velocity_steps[step]
            iq_module.set(velocity_client, velocity_client_entry, new_velocity)

        return 1

    @staticmethod
    def ramp_volts(
        iq_module: IqModule,
        volts_client: str,
        volts_client_entry: str,
        final_volts: float,
        total_time: float,
        time_steps=20,
    ):
        init_volts = iq_module.get_retry("brushless_drive", "drive_volts", retries=5)

        if init_volts == None:
            return 0

        num_steps = int(
            round(1000 * total_time / time_steps)
        )  # 20ms time steps by default

        volts_steps = np.linspace(init_volts, final_volts, num_steps)
        time_steps = np.linspace(0, total_time, num_steps)

        start_time = time.perf_counter()
        for step in range(num_steps):
            while time.perf_counter() < start_time + time_steps[step]:
                pass
            new_voltage = volts_steps[step]
            iq_module.set(volts_client, volts_client_entry, new_voltage)

        return 1

    @staticmethod
    def ramp_volts_slew(iq_module: IqModule, final_volts: float, slew_rate: float):
        init_volts = iq_module.get_retry("brushless_drive", "drive_volts", retries=5)

        if init_volts == None:
            return 0

        time = abs((final_volts - init_volts) / slew_rate)

        success = iq_module.ramp_volts(final_volts, time)
        return success



import time

from iqmotion.client_entries.dictionary_client_entry import AccessType
from iqmotion.clients import client_with_entries
from iqmotion.communication.communicator import Communicator
from iqmotion.custom_errors import IqModuleError
from iqmotion.iq_devices.iq_module_json_parser import IqModuleJsonParser
from iqmotion.message_making.client_with_entries_message import ClientWithEntriesMessage
from iqmotion.message_making.dictionary_message_maker import DictionaryMessageMaker


class IqModule:
    """ IqModule is an asbstract interface class for all the modules made by iq
    """

    _MODULE_FILE_NAME = ""
    _DEFAULT_CONTROL_CLIENT = ""

    def __init__(self, com: Communicator, module_idn=0):
        self._client_dict = {}
        self._com = com
        self._module_idn = module_idn

        module_file_dict = IqModuleJsonParser.parse(self._MODULE_FILE_NAME)
        self._create_clients(module_file_dict)

    def _create_clients(self, module_file_dict: dict):
        # TODO: parse different clients ?
        for client_name in module_file_dict["clients"]:
            self._client_dict[client_name] = client_with_entries.ClientWithEntries(
                client_name, self._module_idn
            )

    def coast(self):
        """ Sends a coast command from the default control client to the module.
        You can check the success with "brushless_drive" "drive_mode".
        """
        self.set(self._DEFAULT_CONTROL_CLIENT, "ctrl_coast")

    def set(self, client_name: str, client_entry_name: str, *args):
        """ Sets a value to the module with a message formed by a client and client entry
        
        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry
            *args -- value(s) to be set
        """
        self._client_and_client_entry_exists(client_name, client_entry_name)

        client = self._client_dict[client_name]
        client_entry = client.client_entries[client_entry_name]

        message_bytes = self._make_message_bytes(client_entry, AccessType.SET, args)

        self._com.send_message(message_bytes)

    def set_verify(
        self,
        client_name: str,
        client_entry_name: str,
        *args,
        time_out=0.1,
        retries=5,
        save=False,
    ):
        """ Sets a value to the module with a message formed by a client and client entry and checks 
        if value was set properly. You can also set "save=True" to save the value if the verify was 
        succesful.

        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry
            *args -- value(s) to be set

        Keyword Arguments:
            time_out {float} -- blocking timeout while verifying the set (s) (default: {0.1})
            retries {int} -- num of times you want to retry verifying the set (default: {5})
            save {bool} -- save the value if verify was successful (default: {False})

        Returns:
            bool -- if the value was correctly set
        """
        verify_range = 0.01
        self.set(client_name, client_entry_name, args)

        get_value = self.get_retry(client_name, client_entry_name, time_out, retries)

        # checks when you do an empty set
        if not args:
            if get_value != 1:
                return False
        else:
            if not self._verify_value(args[0], get_value, verify_range):
                return False

        if save:
            self.save(client_name, client_entry_name)

        return True

    def _verify_value(self, set_value, get_value, range):
        if get_value is None:
            return False
        # some values might loose precision when being saved so check if it's within a 0.01 range
        elif abs(set_value - get_value) >= 0.01:
            return False

        return True

    def get(self, client_name: str, client_entry_name: str, time_out=0.1):
        """ Gets the value define by the client and client entry from the module.

        This call is blocking and will wait until it gets a reply or timeouts.
        
        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry
        
        Keyword Arguments:
            time_out {float} -- blocking timeout while waiting for a reply (s) (default: {0.1})
        
        Returns:
            the reply from the module, None if no reply was available (timeout)
        """
        self.get_async(client_name, client_entry_name)

        max_time = time.perf_counter() + time_out
        while not self.is_fresh(client_name, client_entry_name):
            if time.perf_counter() > max_time:
                return None

            self.update_replies()

        reply = self.get_reply(client_name, client_entry_name)
        return reply

    def get_retry(
        self, client_name: str, client_entry_name: str, time_out=0.1, retries=5
    ):
        """ Sends multiple get requests to the module until a reply comes back or there are no more retries left
        
        Arguments:
            client_name {str} -- name of the client 
            client_entry_name {str} -- name of the client entry
        
        Keyword Arguments:
            time_out {float} -- blocking timeout while waiting for a reply for every retry (s) (default: {0.1})
            retries {int} -- num of times you want to retry sending a get request if nothing came back (default: {5})
        
        Returns:
            the reply from the module, None if no reply was available (timeout and/or max num of retries)
        """
        for _ in range(retries):
            reply = self.get(client_name, client_entry_name, time_out)
            if reply is not None:
                return reply

        return None

    def get_all(self, client_name: str, time_out=0.1):
        """ Gets all the value define by the client (all of its client entries).

        This call is blocking and will wait until it gets a reply or timeouts for every "get" call
        
        Arguments:
            client_name {str} -- name of client
        
        Keyword Arguments:
            time_out {float} --  blocking timeout while waiting for a reply (s) (default: {0.1})
        
        Returns:
            dict -- all the successful replies of the module
        """
        self._client_exists(client_name)
        client = self._client_dict[client_name]

        replies = {}
        for client_entry_name in client.client_entries.keys():
            reply = self.get(client_name, client_entry_name, time_out)
            replies[client_entry_name] = reply

        return replies

    def get_all_retry(self, client_name: str, time_out=0.1, retries=5):
        """[summary]

        Arguments:
            client_name {str} -- name of client

        Keyword Arguments:
            time_out {float} -- blocking timeout while waiting for a reply (s) (default: {0.1})
            retries {int} -- num of times you want to retry sending a get request if nothing came back (default: {5})

        Returns:
            dict -- all the successful replies of the module
        """
        self._client_exists(client_name)
        client = self._client_dict[client_name]

        replies = {}
        for client_entry_name in client.client_entries.keys():
            reply = self.get_retry(client_name, client_entry_name, time_out, retries)
            replies[client_entry_name] = reply

        return replies

    def save(self, client_name: str, client_entry_name: str):
        """ Saves the client and client entry values already set on the module
        
        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry
        """
        self._client_and_client_entry_exists(client_name, client_entry_name)

        client = self._client_dict[client_name]
        client_entry = client.client_entries[client_entry_name]

        message_bytes = self._make_message_bytes(client_entry, AccessType.SAVE, [])

        self._com.send_message(message_bytes)

    def save_all(self, client_name: str):
        """ Saves all the values of the client already set on the module
        
        Arguments:
            client_name {str} -- name of the client
        """
        self._client_exists(client_name)
        client = self._client_dict[client_name]

        for client_entry_name in client.client_entries.keys():
            self.save(client_name, client_entry_name)

    def get_async(self, client_name: str, client_entry_name: str):
        """ Sends a asynchroniously get request to the module

        This call is non blocking, to read the reply you have to call "update reply" or "update replies" and then "get_reply" if that client entry is fresh
        
        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry
        """
        self._client_and_client_entry_exists(client_name, client_entry_name)
        client = self._client_dict[client_name]
        client_entry = client.client_entries[client_entry_name]

        message_bytes = self._make_message_bytes(client_entry, AccessType.GET, [])

        self._com.send_message(message_bytes)

    def flush_input_com_buffer(self):
        """ Flushes the input buffer of your communicator object (such as Serial) and queue
        """
        self._com.flush_input_buffer()

    def update_replies(self):
        """ Reads all the bytes possible in the Communication queue and stores them in the right client entries.
        If this returns False, you might want to run this function again (more bytes in the com queue that was possible to parse at once).

        Returns:
            bool -- True if every bytes were read from the Communication queue, False otherwise
        """
        all_bytes_read = self._com.read_bytes()

        new_message = self._com.extract_message()
        while new_message != None:
            for client in self._client_dict.values():
                client.read_message(new_message)

            new_message = self._com.extract_message()

        return all_bytes_read

    def update_reply(self):
        """ Checks the packet queue (and communication queue if empty) for a new message
        """
        new_message = self._com.extract_message()

        # if no message in the packet queue, read from serial buffer and check again
        if new_message == None:
            self._com.read_bytes()
            new_message = self._com.extract_message()

        if new_message != None:
            for client in self._client_dict.values():
                client.read_message(new_message)

    def is_fresh(self, client_name: str, client_entry_name: str):
        """ Check if the value in the client, client entry is new or not (fresh/not fresh)

        This method is usually called after "update_replies" to check if a client entry got updated
        
        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry
        
        Returns:
            bool -- True if value is fresh, False otherwise
        """
        self._client_and_client_entry_exists(client_name, client_entry_name)
        client = self._client_dict[client_name]
        return client.is_fresh(client_entry_name)

    def get_reply(self, client_name: str, client_entry_name: str):
        """ Reads the reply stored in the client entry

        This method is normally called after checking if the value is fresh with "is_fresh" 
        
        Arguments:
            client_name {str} -- name of the client
            client_entry_name {str} -- name of the client entry

        
        Returns:
            value stored in client entry
        """
        self._client_and_client_entry_exists(client_name, client_entry_name)
        client = self._client_dict[client_name]

        return client.get_reply(client_entry_name)

    def list_clients(self):
        """ Displays all the clients available with the module 
        """
        print("\nClients available from '{0}':\n".format(self._MODULE_FILE_NAME))

        for client_name in self._client_dict.keys():
            print("\t{0}".format(client_name))

        return

    def list_client_entries(self, client_name: str):
        """ Displays all the client entries available for that client

        Args:
            client_name {str} -- name of client
        """
        self._client_exists(client_name)

        # TODO: Make a pretty table with indication of all values
        print("\nEntries available loaded from '{0}': \n".format(client_name))

        client = self._client_dict[client_name]
        for client_entry in client.client_entries.values():
            print(client_entry.data)

        return

    def _client_and_client_entry_exists(self, client_name: str, client_entry_name: str):
        self._client_exists(client_name)

        client = self._client_dict[client_name]

        if client_entry_name not in client.client_entries.keys():
            raise IqModuleError(
                "{0} does not support this client entry: {1}\n".format(
                    client_name, client_entry_name
                )
            )

        return 1

    def _client_exists(self, client_name: str):
        if client_name not in self._client_dict.keys():
            raise IqModuleError(
                "This module does not support this client: {0}\n".format(client_name)
            )
        return True

    def _make_message_bytes(self, client_entry, access_type: AccessType, args):
        message_maker = DictionaryMessageMaker(
            client_entry, self._module_idn, access_type, args
        )

        message = ClientWithEntriesMessage(message_maker)

        return message.make_bytes()



from iqmotion.iq_devices.step_dir_module import StepDirModule
from iqmotion.tests.helpers import MockCommunicator
from iqmotion.iq_devices.common_commands.ramper import Ramper

import pytest
from unittest.mock import patch, Mock, MagicMock, call, PropertyMock


class TestStepDirModule:
    @pytest.fixture
    def mock_communicator(self):
        mock_class = MockCommunicator()
        return mock_class

    def test_module(self, mock_communicator):
        StepDirModule(mock_communicator)

    def test_ramp_velocity(self, mock_communicator):
        Ramper.ramp_velocity = MagicMock()
        module = StepDirModule(mock_communicator)

        velocity_client = "multi_turn_angle_control"
        velocity_client_entry = "ctrl_velocity"
        final_velocity = 5
        total_time = 2
        time_steps = 10

        module.ramp_velocity(final_velocity, total_time, time_steps=time_steps)

        Ramper.ramp_velocity.assert_called_with(
            module,
            velocity_client,
            velocity_client_entry,
            final_velocity,
            total_time,
            time_steps,
        )

    def test_ramp_volts(self, mock_communicator):
        Ramper.ramp_volts = MagicMock()
        module = StepDirModule(mock_communicator)

        volts_client = "multi_turn_angle_control"
        volts_client_entry = "ctrl_volts"
        final_volts = 5
        total_time = 2
        time_steps = 10

        module.ramp_volts(final_volts, total_time, time_steps=time_steps)

        Ramper.ramp_volts.assert_called_with(
            module,
            volts_client,
            volts_client_entry,
            final_volts,
            total_time,
            time_steps,
        )

    def test_ramp_slew(self, mock_communicator):
        Ramper.ramp_volts_slew = MagicMock()
        module = StepDirModule(mock_communicator)

        final_volts = 5
        slew_rate = 2

        module.ramp_volts_slew(final_volts, slew_rate)

        Ramper.ramp_volts_slew.assert_called_with(module, final_volts, slew_rate)
        pass



from abc import ABC, abstractmethod


class ClientEntry(ABC):
    """ClientEntry is an interface class in order to store and access the information expressed by an entry in a client file.
    """

    @abstractmethod
    def read_message(self, msg):
        """ this property is too abstract to understand. """

    @property
    @abstractmethod
    def fresh(self):
        """ this property is too abstract to understand. """

    @property
    @abstractmethod
    def value(self):
        """ this property is too abstract to understand. """

    @value.setter
    @abstractmethod
    def value(self, val):
        """ this property is too abstract to understand. """

    @property
    @abstractmethod
    def data(self):
        """ this property is too abstract to understand. """



import iqmotion as iq

import pytest
import time


class TestGeneric:
    @pytest.fixture
    def iq_module(self, hardware_type, usb_handle):
        hardware_types = {
            "speed": iq.SpeedModule,
            "servo": iq.ServoModule,
            "step_dir": iq.StepDirModule,
            "pulsing": iq.PulsingModule,
        }

        if hardware_type not in hardware_types.keys():
            print("THIS TYPE IS NOT HANDLED IN TEST GENERING")
            exit()

        com = iq.SerialCommunicator(usb_handle)
        iq_module = hardware_types[hardware_type](com, 0)

        # timeout for speed modules
        if hardware_type == "speed":
            iq_module.set("propeller_motor_control", "timeout", 10)

        return iq_module

    def test_get_all(self, iq_module):
        empty_entries = ["reboot_program", "reboot_boot_loader"]

        results = iq_module.get_all("system_control")
        for key, value in results.items():
            if key not in empty_entries:
                assert value != None

    def test_get_overload(self, iq_module):
        for _ in range(50):
            result = iq_module.get("system_control", "uid1")
            assert result != None

    def test_get_async_overload(self, iq_module):
        angle_samples = []
        sampling_time = 5

        start_time = time.perf_counter()
        while time.perf_counter() < start_time + sampling_time:
            iq_module.get_async("brushless_drive", "obs_angle")
            all_bytes_read = iq_module.update_replies()
            while not all_bytes_read:
                all_bytes_read = iq_module.update_replies()

            if iq_module.is_fresh("brushless_drive", "obs_angle"):
                current_angle = iq_module.get_reply("brushless_drive", "obs_angle")
                angle_samples.append(current_angle)

        assert None not in angle_samples

    def test_set_verify(self, iq_module):
        orginal_kv = iq_module.get("brushless_drive", "motor_Kv")

        success = iq_module.set_verify("brushless_drive", "motor_Kv", orginal_kv + 1)
        assert success == True

        iq_module.set_verify("brushless_drive", "motor_Kv", orginal_kv, save=True)

    def test_coast(self, iq_module):
        iq_module.set("brushless_drive", "drive_volts", 1)
        time.sleep(0.5)
        drive_mode = iq_module.get_retry("brushless_drive", "drive_mode")

        assert drive_mode != 5

        iq_module.set("brushless_drive", "drive_coast")
        drive_mode = iq_module.get_retry("brushless_drive", "drive_mode")
        assert drive_mode == 5



from iqmotion.message_making.message_maker import MessageMaker

from abc import ABC, abstractmethod


class Message(ABC):
    """ Message is the interface class used by the Specialiizations in a bridge pattern.
    Each implemenation should correspond to a different type of Clien implementation
    """

    def __init__(self, message_maker: MessageMaker):
        self._message_maker = message_maker

    @abstractmethod
    def make_bytes(self):
        """ this property is too abstract to understand. """



from iqmotion.communication.packet_state import PacketState
from iqmotion.communication.circular_queue import CircularQueue
from iqmotion.communication.crc import Crc

from iqmotion.custom_errors import PacketStateError

import copy


class SerialStartState(PacketState):
    """ SerialStartState defines the start state of a serial packet.
    It will parse the circular byte queue until it finds its start byte (0x55)
    """

    _START_BYTE = 0x55

    def __init__(
        self,
        byte_queue: CircularQueue,
        start_index: int = 0,
        parse_index: int = 0,
        packet_len: int = 0,
    ):
        super().__init__(byte_queue, start_index, parse_index, packet_len)
        self._end_index = len(byte_queue)

    def parse(self):
        """ popleft every byte in the circular queue until it finds the start byte (0x55)
        """
        parse_index = self._start_index
        for _ in range(parse_index, self._end_index):
            if self._byte_queue[parse_index] == self._START_BYTE:
                self._parse_succesful = 1
                self._parse_index = parse_index
                return
            else:
                self._byte_queue.popleft()

        self._is_done = 1
        self._succesful = 0
        self._end_index = self._parse_index + 1

    def find_next_state(self):
        """ Finds next state depending on the success of parse

        Returns:
            next_state (SerialLenState): if parsing was successful
            next_state (SerialStartState): if parsing not successful
        """
        if self._parse_succesful:
            return SerialLenState(
                self._byte_queue, self._start_index, self._parse_index, self._packet_len
            )
        else:
            return self


class SerialLenState(PacketState):
    """ SerialLenState defines the Len state of a serial packet.
    It will parse the circular byte queue to find the Len Byte and store it in memory.
    Len Byte cannot be more than 250
    """

    _MAX_PACKET_SIZE = 255

    def __init__(
        self,
        byte_queue: CircularQueue,
        start_index: int = 0,
        parse_index: int = 0,
        packet_len: int = 0,
    ):
        super().__init__(byte_queue, start_index, parse_index, packet_len)

    def parse(self):
        """ Find len byte and stores it in memory, if Len Byte > 250, raise an error

        Raise:
            PacketStateError: Packet overflow, message is bigger than 256 bytes
        """
        try:
            self._packet_len = self._byte_queue[self._parse_index + 1]
            if self._packet_len <= self._MAX_PACKET_SIZE:
                self._parse_index += 1
                self._parse_succesful = 1
            else:
                raise PacketStateError(
                    "Packet overflow, message is bigger than 256 bytes"
                )
        except IndexError:
            self._is_done = 1
            self._end_index = self._parse_index + 1
            self._parse_succesful = 0
        return

    def find_next_state(self):
        """ Finds next state depending on the success of parse 

        Returns:
            next_state (SerialTypeState): if parsing was successful
            next_state (SerialLenState): if parsing not successful
        """
        if self._parse_succesful:
            return SerialTypeState(
                self._byte_queue, self._start_index, self._parse_index, self._packet_len
            )
        else:
            return self


class SerialTypeState(PacketState):
    """ SerialTypeState defines the Type state of a serial packet.
    It will parse the circular byte queue to find the Type Byte.
    If the parsing was succesfull it will return SerialPayloadState if Len of payload > 0,
    otherwise it will skip SerialPayloadState and return SericalCrcState
    """

    def __init__(
        self,
        byte_queue: CircularQueue,
        start_index: int = 0,
        parse_index: int = 0,
        packet_len: int = 0,
    ):
        super().__init__(byte_queue, start_index, parse_index, packet_len)

    def parse(self):
        """ Find Type byte
        """

        try:
            self._byte_queue[self._parse_index + 1]
            self._parse_index += 1
            self._parse_succesful = 1
            return
        except IndexError:
            self._is_done = 1
            self._end_index = self._parse_index + 1
            self._parse_succesful = 0
        return

    def find_next_state(self):
        """ Finds next state depending on the success of parse 

        Returns:
            next_state (SerialPayloadState): if parsing was successful and packet_len > 0
            next_state (SerialCrcState): if parsing was successful and packet_len = 0
            next_state (SerialTypeState): if parsing not successful
        """
        if self._parse_succesful:
            if self._packet_len > 0:
                return SerialPayloadState(
                    self._byte_queue,
                    self._start_index,
                    self._parse_index,
                    self._packet_len,
                )
            else:
                return SerialCrcState(
                    self._byte_queue,
                    self._start_index,
                    self._parse_index,
                    self._packet_len,
                )
        else:
            return self


class SerialPayloadState(PacketState):
    """ SerialPayloadState defines the Payload state of a serial packet.
    It will parse the circular byte queue to find all the payload bytes of the packet.
    """

    def __init__(
        self,
        byte_queue: CircularQueue,
        start_index: int = 0,
        parse_index: int = 0,
        packet_len: int = 0,
    ):
        super().__init__(byte_queue, start_index, parse_index, packet_len)

    def parse(self):
        """ Find Paylod bytes
        """
        try:
            for ind in range(
                self._parse_index + 1, self._parse_index + 1 + self._packet_len
            ):
                # check if there is the right amount of bytes,
                self._byte_queue[ind]
            self._parse_index += self._packet_len
            self._parse_succesful = 1
            return

        except IndexError:
            self._is_done = 1
            self._end_index = self._parse_index + 1
            self._parse_succesful = 0

        return

    def find_next_state(self):
        """ Finds next state depending on the success of parse 

        Returns:
            next_state (SerialCrcState): if parsing was successful
            next_state (SerialPayloadState): if parsing not successful
        """
        if self._parse_succesful:
            return SerialCrcState(
                self._byte_queue, self._start_index, self._parse_index, self._packet_len
            )
        else:
            return self


class SerialCrcState(PacketState):
    """ SerialCrcState defines the Crc state of a serial packet.
    It will check the current packet for the Crc bytes and verify that they are correct.
    If the Crc is correct, stores the message in its message @property
    """

    def __init__(
        self,
        byte_queue: CircularQueue,
        start_index: int = 0,
        parse_index: int = 0,
        packet_len: int = 0,
    ):
        super().__init__(byte_queue, start_index, parse_index, packet_len)

    def parse(self):
        """ Find Crc bytes and check if they are correct.
        If correct, store message in its message @property
        """
        try:
            crcl = self._byte_queue[self._parse_index + 1]
            crch = self._byte_queue[self._parse_index + 2]
            crc = (crch << 8) | crcl

            msg = self._extract_packet_message()
            expected_crc = Crc.make_crc(msg)

            if crc == expected_crc:
                self._message = copy.deepcopy(msg[1:])
                self._end_index = self._parse_index + 3
                self._is_done = 1
                self._succesful = 1
            else:
                self._is_done = 1
                self._end_index = self._parse_index + 3
                self._pop_message_bytes()
                self._succesful = 0

        except IndexError:
            self._is_done = 1
            self._end_index = self._parse_index + 1
            self._succesful = 0

        return

    def find_next_state(self):
        """ Finds next state depending on the success of parse 

        Returns:
            next_state (SerialCrcState): if parsing was successful
        """
        return self

    def _extract_packet_message(self):
        msg_start = self._start_index + 1
        # the two is the Len and Type of the data
        msg_end = msg_start + self._parse_index
        msg = bytearray(self._byte_queue[msg_start:msg_end])
        return msg

    def _pop_message_bytes(self):
        num_bytes_to_pop = self._end_index - self._start_index
        for _ in range(num_bytes_to_pop):
            self._byte_queue.popleft()



class CircularQueue:
    """ A fix size circular queue implemented with a list() for O(1) memory movement costs anywhere in the queue
        The queue is FIFO if you use the __iter__ method and/or popleft(), but can be LIFO with pop()
    """

    def __init__(self, maxlen: int):
        """ Create an empty circular queue of ints of size "size"

        Args:
            size (int): Size of the circular queue
        """

        self._queue = [None for i in range(maxlen)]
        self._front = -1
        self._rear = -1
        self._count = 0
        self._maxlen = maxlen

    @classmethod
    def from_iterable(cls, iterable: iter):
        """ Create an circular queue from an iterable with size len(iterable)

        Args:
            iterable (iter): The iterable the queue will copy
        """
        iterable_len = len(iterable)

        new_queue = CircularQueue(iterable_len)
        new_queue.extend(iterable)

        return new_queue

    def __iter__(self):
        """ Peek iteratively over the queue in a FIFO way, this method does NOT pop the element
        """
        if not self.is_empty:

            front_index = self._front
            for _ in range(self._count):
                peeked_element = self._queue[front_index]
                front_index = (front_index + 1) % self._maxlen

                yield peeked_element

    def __len__(self):
        return self._count

    def __str__(self):
        list_from_queue = self._queue[self._front : self._rear + 1]
        return list_from_queue.__str__()

    def __getitem__(self, key):
        """ Slicing multiple elements is not fast, this will copy the queue as a list and slice this list.
            Slicing one element is alot faster, equivalent to a "peek" but for any index position
        """
        if type(key) == slice:
            return self._parse_queue_from_slice(key)

        self._check_key_in_range(key)

        actual_index = self._calculate_actual_queue_index(key)

        return self._queue[actual_index]

    def _parse_queue_from_slice(self, slice):
        queue_as_list = []
        for item in self:
            queue_as_list.append(item)

        return queue_as_list[slice]

    def _check_key_in_range(self, key):
        if key >= 0:
            if key > self._count - 1:
                raise IndexError("IndexError: queue index out of range")
        else:
            if abs(key) > self._count:
                raise IndexError("IndexError: queue index out of range")

    def _calculate_actual_queue_index(self, key):
        if key < 0:
            wanted_index = self._count + key
        else:
            wanted_index = key

        actual_index = self._front + wanted_index
        if actual_index > self._maxlen - 1:
            actual_index = actual_index % self._count

        return actual_index

    def append(self, x):
        """ Add an item to the end of the list.

        Args:
            x: Item you want to append.

        Returns:
            true: if succesful
            false: if circular queue is full
        """
        if self.is_full:
            return 0

        # condition for empty queue
        elif self.is_empty:
            self._front = 0
            self._rear = 0
        else:
            # next position of rear
            self._rear = (self._rear + 1) % self._maxlen

        self._count += 1
        self._queue[self._rear] = x

        return 1

    def extend(self, iterable: iter):
        """ Extend the list by appending all the items from the iterable.

        Args:
            iterable (iter): Iterable object you want to add to your queue.

        Returns:
            true: if succesful
            false: if circular queue is full
        """
        for x in iterable:
            if not self.append(x):
                return 0

        return 1

    def pop(self):
        """ Remove and return an element from the right side of the circular queue.

        Returns:
            None: if queue is empty
            poped_element: element from queue that got poped
        """
        if self.is_empty:
            return None

        poped_element = self._queue[self._rear]

        # condition for only one element
        if self._front == self._rear:
            self._front = -1
            self._rear = -1
        else:
            self._rear = (self._rear - 1) % self._maxlen

        self._count -= 1
        return poped_element

    def popleft(self):
        """ Remove and return an element from the left side of the circular queue.

        Returns:
            None: if queue is empty
            poped_element: element from queue that got poped
        """
        if self.is_empty:
            return None

        poped_element = self._queue[self._front]

        # condition for only one element
        if self._front == self._rear:
            self._front = -1
            self._rear = -1
        else:
            self._front = (self._front + 1) % self._maxlen

        self._count -= 1
        return poped_element

    def clear(self):
        """ Remove all elements from the circular queue.
        """
        self._front = -1
        self._rear = -1
        self._count = 0

    def copy(self):
        """ copy of the circular queue.

        Return:
            (CircularQueue): copy of circular queue
        """
        cq = CircularQueue.from_iterable(self)
        return cq

    @property
    def maxlen(self):
        """ Maximum size of the circular queue.

        Return:
            Maximum size of the circular queue.
        """
        return self._maxlen

    @property
    def is_full(self):
        """ Check if queue is full

        Returns:
            true: if queue is full
        """
        # condition if queue is full
        if (self._rear + 1) % self._maxlen == self._front:
            return 1
        else:
            return 0

    @property
    def is_empty(self):
        """ Check if queue is empty

        Returns:
            true: if queue is empty
        """
        if self._front == -1:
            return 1
        else:
            return 0



from iqmotion.communication.serial_packet_queue import SerialPacketQueue
from iqmotion.communication.crc import Crc
from iqmotion.custom_errors import PacketQueueError

from iqmotion.tests.helpers import make_fake_packet

import pytest


class TestSerialPacketQueue:
    @pytest.fixture(params=[[1], [1, 2, 3, 4, 5]])
    def serial_packet_queue(self, request):
        if len(request.param) != 1:
            packet = make_fake_packet(request.param)
        else:
            packet = bytes(request.param)

        spq = SerialPacketQueue()
        spq.put_bytes(packet)

        return [spq, packet]

    def test__len__(self, serial_packet_queue):
        spq = serial_packet_queue[0]
        data = serial_packet_queue[1]

        assert len(data) == len(spq)

    def test_is_empty(self, serial_packet_queue):
        spq = serial_packet_queue[0]

        assert False == spq.is_empty

        spq.drop_packet()
        assert True == spq.is_empty

    def test_put_bytes(self):
        data = [i for i in range((255 + 5) * 2)]
        spq = SerialPacketQueue()

        with pytest.raises(PacketQueueError) as err:
            assert spq.put_bytes(data) == 1
            spq.put_bytes(1)

        err_str = err.value.message
        assert "PACKET QUEUE ERROR: Byte Queue Overflow\n" == err_str

        with pytest.raises(PacketQueueError) as err:
            assert spq.put_bytes(data) == 1
            spq.put_bytes([1, 2, 3])

        err_str = err.value.message
        assert "PACKET QUEUE ERROR: Byte Queue Overflow\n" == err_str

    def test_peek(self, serial_packet_queue):
        spq = serial_packet_queue[0]
        data = serial_packet_queue[1]

        if type(data) == bytes:
            assert None == spq.peek()
        else:
            expected_data = data[2:-2]
            assert expected_data == spq.peek()

    def test_peek_empty(self):
        spq = SerialPacketQueue()

        with pytest.raises(PacketQueueError) as err:
            spq.peek()

        err_str = err.value.message
        assert "PACKET QUEUE ERROR: Serial packet queue is empty\n" == err_str

    def test_drop_packet(self, serial_packet_queue):
        spq = serial_packet_queue[0]
        data = serial_packet_queue[1]

        spq.peek()

        if type(data) == bytes:
            assert False == spq.drop_packet()
        else:
            assert True == spq.drop_packet()



class ClientError(Exception):
    err_type = "CLIENT ERROR"

    def __init__(self, err_descriptor):
        super().__init__(err_descriptor)
        self.message = self.err_type + ": " + err_descriptor + "\n"


class CommunicationError(Exception):
    err_type = "COMMUNICATION ERROR"

    def __init__(self, err_descriptor):
        super().__init__(err_descriptor)
        self.message = self.err_type + ": " + err_descriptor + "\n"


class PacketQueueError(Exception):
    err_type = "PACKET QUEUE ERROR"

    def __init__(self, err_descriptor):
        super().__init__(err_descriptor)
        self.message = self.err_type + ": " + err_descriptor + "\n"


class PacketStateError(Exception):
    err_type = "PACKET STATE ERROR"

    def __init__(self, err_descriptor):
        super().__init__(err_descriptor)
        self.message = self.err_type + ": " + err_descriptor + "\n"


class IqModuleError(Exception):
    err_type = "IQ MODULE ERROR"

    def __init__(self, err_descriptor):
        super().__init__(err_descriptor)
        self.message = self.err_type + ": " + err_descriptor + "\n"


class MessageMakerError(Exception):
    err_type = "MESSAGE MAKER ERROR"

    def __init__(self, err_descriptor):
        super().__init__(err_descriptor)
        self.message = self.err_type + ": " + err_descriptor + "\n"
